from typing import Optional

from aioqbt._paramdict import ParamDict
from aioqbt.api.types import SyncMainData, SyncTorrentPeers
from aioqbt.bittorrent import InfoHash
from aioqbt.client import APIGroup

__all__ = ("SyncAPI",)


# Result from sync/maindata and sync/torrentPeers are "differenced":
# - "rid" key always exist
# - "full_update" is either true or missing
# - other attributes are missing if unchanged
# - complex/nested structure, eg, dict and list, are also "differenced"
# Default values are assigned to the missing attributes of the result:
# - "full_update" is either true or false
# For SyncMainData,
# - "torrents" and "categories" attributes are empty dict when missing
# - "*_remove" attributes are empty list when missing
# For TorrentPeers,
#


class SyncAPI(APIGroup):
    """
    Sync APIs.

    In Sync APIs, changes between requests are returned in dict-like objects.
    Keys may be omitted if their values are unchanged.

    """

    async def maindata(
        self,
        rid: Optional[int] = None,
    ) -> SyncMainData:
        """
        Obtain sync data.

        ``rid`` in the previous sync data may be passed to the ``rid`` argument
        to obtain a difference update.

        If ``full_update=True`` in the resultant object, the data is a full update.
        Otherwise, the data only contains changes since the last sync request.

        """
        params = ParamDict()
        params.optional_int("rid", rid)

        return await self._request_mapped_object(
            SyncMainData,
            "GET",
            "sync/maindata",
            params=params,
        )

    async def torrent_peers(
        self,
        hash: InfoHash,
        rid: Optional[int] = None,
    ) -> SyncTorrentPeers:
        """
        Obtain peers for a torrent.

        ``rid`` and ``full_update`` share similar meanings in :meth:`.maindata`.

        """
        params = ParamDict.with_hash(hash)
        params.optional_int("rid", rid)

        return await self._request_mapped_object(
            SyncTorrentPeers,
            "GET",
            "sync/torrentPeers",
            params=params,
        )



from aioqbt import exc
from aioqbt.client import APIClient, APIGroup

__all__ = ("AuthAPI",)


async def _auth_login(client: APIClient, username: str, password: str) -> None:
    resp = await client.request(
        "POST",
        "auth/login",
        data={
            "username": str(username),
            "password": str(password),
        },
    )

    async with resp:
        res = await resp.read()

        if res != b"Ok.":
            ex = exc.LoginError.from_response(resp)
            ex.message = res.decode("utf-8")
            raise ex


class AuthAPI(APIGroup):
    """
    API methods under ``auth``.
    """

    async def login(self, username: str, password: str) -> None:
        return await _auth_login(self._client(), username, password)

    async def logout(self) -> None:
        # Seem that logout always succeed
        await self._request_text(
            "POST",
            "auth/logout",
        )



import sys
from enum import Enum
from typing import Callable, Generic, Optional, Type, TypeVar, Union, overload

T = TypeVar("T")
R = TypeVar("R")
E = TypeVar("E", bound=Enum)


@overload
def repr_enum(cls: None) -> Callable[[Type[E]], Type[E]]:
    pass


@overload
def repr_enum(cls: Type[E]) -> Type[E]:
    pass


def repr_enum(
    cls: Optional[Type[E]] = None,
) -> Union[Callable[[Type[E]], Type[E]], Type[E]]:
    """
    Preserve str() and format() behavior of enum types from underlying
    data type.

    Decorate an enum subclass mixed with a data type such that str() and
    format() return the equivalent result similar to the function
    directly operated on the objects of the data type.

    Mimic ReprEnum on Python 3.11.
    See https://github.com/python/cpython/issues/85955
    """

    def decorate(cls: Type[E]) -> Type[E]:
        if not issubclass(cls, Enum):
            raise TypeError(f"require subclass of Enum instead of {cls!r}")

        mro = tuple(cls.__mro__)

        # Find the first enum subclass in mro except cls itself
        enum_idx = next(i for i, s in enumerate(mro[1:], 1) if issubclass(s, Enum))
        enum_type = mro[enum_idx]
        mixin_types = mro[1:enum_idx]

        member_type = cls._member_type_  # type: ignore
        if member_type is object or member_type not in mixin_types:
            raise TypeError("repr_enum() only decorate enum types mixed with a data type")

        # If both cls and enum_type refer to the same method,
        # assume that the method is set by EnumMeta
        # even though the method may be explicitly defined in mixins

        if cls.__format__ is enum_type.__format__:  # type: ignore[comparison-overlap]
            cls.__format__ = member_type.__format__  # type: ignore[method-assign]

        if cls.__str__ is enum_type.__str__:  # type: ignore[comparison-overlap]
            # object.__str__ delegate to member_type.__repr__
            if member_type.__str__ is object.__str__:
                str_method = member_type.__repr__
            else:
                str_method = member_type.__str__

            cls.__str__ = str_method  # type: ignore[method-assign]

        return cls

    if cls is None:
        return decorate  # pragma: no cover
    else:
        return decorate(cls)


if sys.version_info >= (3, 11):  # pragma: no cover
    from enum import IntEnum, StrEnum

    # avoid unused references
    assert True, (StrEnum, IntEnum)
else:

    @repr_enum
    class StrEnum(str, Enum):
        pass

    @repr_enum
    class IntEnum(int, Enum):
        pass


if sys.version_info >= (3, 8):
    from functools import cached_property

    assert True, cached_property
else:  # pragma: no cover

    class cached_property(Generic[T, R]):
        """
        Modified from the functools.cached_property implementation on Python 3.8
        """

        def __init__(self, func: Callable[[T], R]) -> None:
            self.func = func
            self.name: Optional[str] = None
            self.__doc__ = func.__doc__

        def __set_name__(self, owner: T, name: str) -> None:
            if self.name is None:
                self.name = name
            elif self.name != name:
                raise TypeError(
                    f"cached_property cannot be associated with two different names"
                    f" ({self.name!r} and {name!r})"
                )

        @overload
        def __get__(
            self,
            instance: None,
            owner: Optional[Type[T]] = ...,
        ) -> "cached_property[T, R]":
            ...

        @overload
        def __get__(
            self,
            instance: T,
            owner: Optional[Type[T]] = ...,
        ) -> R:
            ...

        def __get__(
            self,
            instance: Optional[T],
            owner: Optional[Type[T]] = None,
        ) -> Union["cached_property[T, R]", R]:
            if instance is None:
                return self

            if self.name is None:
                raise TypeError("cached_property must be associated with a name")

            try:
                cache = instance.__dict__
            except AttributeError:
                raise TypeError(
                    "cached_property require __dict__ attribute to cache value"
                ) from None

            try:
                value = cache[self.name]
            except KeyError:
                value = self.func(instance)
                cache[self.name] = value
            return value  # type: ignore[no-any-return]



from typing import Iterable, Union

from typing_extensions import Literal

InfoHash = Union[str, bytes]
InfoHashes = Iterable[InfoHash]
InfoHashesOrAll = Union[InfoHashes, Literal["all"]]


def get_info_hash(obj: InfoHash) -> str:
    """Get info hash"""

    if isinstance(obj, (bytes, bytearray, memoryview)):
        obj = obj.hex()

    if len(obj) != 40 and len(obj) != 64:
        raise ValueError("info hash is a hexadecimal string of 40 or 60 characters")

    try:
        int(obj, 16)
    except ValueError:
        raise ValueError("info hash is a hexadecimal string of 40 or 60 characters")

    return obj


def _info_hash_str(value: InfoHash) -> str:
    """Convert info hash to str"""

    if isinstance(value, bytes):
        return value.hex()

    return value
